#ifndef __RT_LIB_RT_MESSAGE_H__
#define __RT_LIB_RT_MESSAGE_H__

#include <ostream>
#include <cstddef>
#ifndef _MSC_VER
#include <inttypes.h>
#else
#if _MSC_VER >= 1600
#include <cstdint>
#else
typedef unsigned __int64 uint64_t;
#endif  //  _MSC_VER >= 1600
#endif // !_MSC_VER
#include <fire/private/prefix.h>
#include <fire/detail/RT_FidField.h>

namespace FIRE_NAMESPACE {

class RT_API RT_Message
{
  public:

  class const_iterator;  // Acts like a pointer to an RT_FidField

  RT_Message();
  RT_Message(const RT_Message &other);
  RT_Message &operator=(const RT_Message &other);
  ~RT_Message();

  // set functions (Note: len param should not include NULL termination)
  RT_Message& set_key          (const char *key_val, int key_len);
  RT_Message& set_permissions  (const char *perm, int perm_len);
  RT_Message& set_stale        (bool is_stale, const char *rsn, int rsn_len);
  RT_Message& set_error        (rt_errno err, const char *error_descr, int len);

  RT_Message& set_aggregatable (bool is_aggregatable = true);
  RT_Message& set_aggregated   (bool is_aggregated = true);
  RT_Message& set_response_flag(bool is_initial_response = true);
  RT_Message& set_complete     (bool is_complete = true);
  RT_Message& set_close        (bool is_closed   = true);

  // set functions that take a null-terminated string (no len param)
  RT_Message& set_key          (const char *key_val);
  RT_Message& set_permissions  (const char *perm);
  RT_Message& set_stale        (bool is_stale, const char *rsn);
  RT_Message& set_error        (rt_errno err,  const char *error_descr);
  
  // get functions return a null-terminated string, len is an optional out param
  const char *get_key              (int *key_len   = NULL) const;
  const char *get_permissions      (int *perms_len = NULL) const;
  const char *get_stale_reason     (int *rsn_klen  = NULL) const;
  const char *get_error_description(int *descr_len = NULL) const;
  rt_errno    get_error()                                  const;

  bool is_response()     const;
  bool is_update()       const;
  bool is_stale()        const;
  bool is_error()        const;
  bool is_closed()       const;
  bool is_aggregatable() const;
  bool is_aggregated()   const;
  bool is_permissioned() const;
  bool is_complete()     const;

  // Takes a null-terminated string (no len param)
  rt_errno set(int fid, const char *value);
  // Takes a string with len (Note: len param should not include NULL termination)
  rt_errno set(int fid, const char *value, int value_len);
  rt_errno set(int fid, const RT_Field &fld);
  rt_errno set(const RT_FidField &fld);
  
  // Takes a null-terminated string (no len param)
  rt_errno append(int fid, const char *value);
  // Takes a string with len (Note: len param should not include NULL termination)
  rt_errno append(int fid, const char *value, int  val_size);
  rt_errno append(int fid, const RT_Field &fld);
  rt_errno append(const RT_FidField &fld);
  
  // Doesn't return a null-terminated pointer, MUST pass in a pointer for len
  const char *get(int fid, int *val_size)      const;
  RT_Field    get(int fid)                     const;
  
  const_iterator begin() const;
  const_iterator end()   const;
  
  // Doesn't return a null-terminated pointer, MUST pass in a pointer for len
  const char *get_by_idx(int idx, int *fid, int *val_size)  const;
  RT_FidField get_by_idx(int idx)                           const;
  
  int  count()         const;
  bool exists(int fid) const;
  bool empty()         const;
  void erase(int fid);
  void clear();

  // Sets everything to 0
  void reset();

  //  returns time of last deserialization of the message, in UTC
  uint64_t get_msg_time() const;

  //
  //private: Clients should not use
  //
  enum MessageFlags {
    MSG_ERROR            = 0x80,
    MSG_CLOSED           = 0x40,
    MSG_COMPLETE         = 0x20,
    MSG_STALE            = 0x10,
    MSG_PERMISSIONS_SET  = 0x08,
    MSG_NON_AGGREGATABLE = 0x04,
    MSG_INITIAL          = 0x02,
    MSG_AGGREGATED       = 0x01
  };

  unsigned int      size() const;
  rt_errno          serialize(void *dest, unsigned int *bytes_written) const;
  rt_errno          deserialize(const void *src, unsigned int src_len);
  int               get_tag()        const;
  const RT_Message& set_tag(int tag) const;
  RT_Message(class RT_MessageRep *b_rep) : mb_rep(b_rep), tag(0),
              error(RT_NO_ERROR), flags(0) {}
  class RT_MessageRep *mb_rep;

  private:

    mutable int     tag;
    rt_errno        error;
    unsigned char   flags; 
};

//
// The rest is implementation
//
inline const RT_Message& RT_Message::set_tag(int t) const
{ tag = t; return *this; }

inline RT_Message& RT_Message::set_aggregatable(bool is_aggregatable)
{ if (!is_aggregatable) flags |= MSG_NON_AGGREGATABLE;
  else flags &= ~(MSG_NON_AGGREGATABLE);
  return *this;
}

inline RT_Message& RT_Message::set_aggregated(bool is_aggregated)
{ if (is_aggregated) flags |= MSG_AGGREGATED; else flags &= ~(MSG_AGGREGATED);
  return *this;
}

inline RT_Message& RT_Message::set_complete(bool complete)
{ if (complete) flags |= MSG_COMPLETE; else flags &= ~(MSG_COMPLETE);
  return *this;
}

inline RT_Message& RT_Message::set_response_flag(bool response)
{ if (response) flags |= MSG_INITIAL; else flags &= ~(MSG_INITIAL);
  return *this;
}

inline RT_Message& RT_Message::set_close(bool closed)
{ if (closed) flags |= MSG_CLOSED; else flags &= ~(MSG_CLOSED);
  return *this;
}

//
// get functions
//
inline int RT_Message::get_tag() const
{ return tag; }

inline rt_errno RT_Message::get_error() const
{ return error;  }

inline bool RT_Message::is_aggregatable() const
{ return (flags & MSG_NON_AGGREGATABLE) ? false : true; }

inline bool RT_Message::is_aggregated() const
{ return (flags & MSG_AGGREGATED) ? true : false; }

inline bool RT_Message::is_permissioned() const
{ return (flags & MSG_PERMISSIONS_SET) ? true : false; }

inline bool RT_Message::is_response() const
{ return (flags & MSG_INITIAL) ? true : false; }

inline bool RT_Message::is_update() const
{ return !is_response(); }

inline bool RT_Message::is_stale() const
{ return (flags & MSG_STALE) ? true : false; }

inline bool RT_Message::is_error() const
{ return (flags & MSG_ERROR) ? true : false; }

inline bool RT_Message::is_closed() const
{ return (flags & MSG_CLOSED) ? true : false; }

inline bool RT_Message::is_complete() const
{ return (flags & MSG_COMPLETE) ? true : false; }

class RT_Message::const_iterator
{
  public:
  typedef RT_FidField value_type;

  const_iterator() : msg(NULL), idx(255) {}
  const_iterator(const RT_Message *r, int i) : msg(r), idx(i) { fetch(); }
  bool operator!=(const const_iterator &rhs) const { return rhs.idx != idx; }
  bool operator==(const const_iterator &rhs) const { return rhs.idx == idx; }

  const_iterator operator++()
  { ++idx; fetch(); return *this; }

  const_iterator operator++(int)
  { const_iterator tmp(*this); ++idx; fetch(); return tmp; }

  const value_type &operator*()  const { return field; }
  const value_type *operator->() const { return &field; }

  private:
  const RT_Message *msg;
  value_type  field;
  int         idx;
  void fetch()
  { field.data.ptr = msg->get_by_idx(idx,&field.fid, &field.data.size); }
};

inline RT_Message::const_iterator RT_Message::begin() const
{ return const_iterator(this, 0);          }
inline RT_Message::const_iterator RT_Message::end()   const
{ return const_iterator(this, count()); }

static inline std::ostream& operator<<(std::ostream &os, const RT_Message &msg)
{
  os << "  T:" << msg.get_tag() << " K:" << msg.get_key()
     << " E:" << msg.get_error() << " Flags";

  if (msg.is_response())       os << ":RSP";
  if (msg.is_update())         os << ":UPD";
  if (msg.is_error())          os << ":ERR";
  if (msg.is_closed())         os << ":FIN";
  if (msg.is_complete())       os << ":CMP";
  if (msg.is_stale())          os << ":STL";
  if (msg.is_permissioned())   os << ":PER";
  if (msg.is_aggregatable())   os << ":AGB";
  if (msg.is_aggregated())     os << ":AGD";

  os << "\n";
  if (msg.is_error()) os << "  error desc-> "
                         << msg.get_error_description() << '\n';
  if (msg.is_stale()) os << "  stale reas-> " << msg.get_stale_reason() << '\n';
  if (msg.is_permissioned()) os << "  permissions-> "
                                << msg.get_permissions() << '\n';

  os << "  NumFids = " << msg.count() << " Size = " << msg.size() << '\n';
  for (RT_Message::const_iterator it=msg.begin(); it != msg.end(); ++it)
     os << "     " << *it << '\n';
  return os;
}

inline rt_errno RT_Message::set(int fid, const RT_Field &fld)
{ return (fld.ptr) ? set(fid, fld.ptr, fld.size) : (erase(fid),RT_NO_ERROR); }

inline rt_errno RT_Message::set(const RT_FidField &fld)
{ return (fld.data.ptr) ? set(fld.fid, fld.data.ptr, fld.data.size) :
                             (erase(fld.fid), RT_NO_ERROR);
}

inline rt_errno RT_Message::append(int fid, const RT_Field &fld)
{ return (fld.ptr) ? append(fid, fld.ptr, fld.size) : RT_NO_ERROR; }

inline rt_errno RT_Message::append(const RT_FidField &fld)
{ return (fld.data.ptr) ? append(fld.fid, fld.data.ptr, fld.data.size) :
                          RT_NO_ERROR;
}

inline RT_Field RT_Message::get(int fid) const
{  RT_Field f; f.ptr = get(fid, &f.size); return f; }

inline RT_FidField RT_Message::get_by_idx(int idx) const
{ RT_FidField f; f.data.ptr = get_by_idx(idx, &f.fid, &f.data.size); return f; }

} // namespace

#endif
