#ifndef __RT_LIB_RT_RECORD_H__
#define __RT_LIB_RT_RECORD_H__

#include <ostream>
#include <fire/detail/RT_FidField.h>
#include <fire/private/prefix.h>

class RouterClient;
class RouterServer;

namespace FIRE_NAMESPACE {

class RT_API RT_Record
{
  public:

  class const_iterator;  // Acts like a pointer to an RT_FidField

  RT_Record(const class RT_Message &msg);
  RT_Record(const RT_Record  &other);
  RT_Record& operator=(const RT_Record &other);
  ~RT_Record();

  // Doesn't return a null-terminated string, MUST pass in a pointer for len
  const char *get_key(int *key_len) const;
  RT_Field    get_key()             const; 

  rt_errno set(int fid, const char *value, int value_len);
  rt_errno set(int fid, const RT_Field &fld);
  rt_errno set(const RT_FidField &fld);

  // Doesn't return a null-terminated pointer, MUST pass in a pointer for len
  const char *get(int fid, int *value_size) const;
  RT_Field    get(int fid)                  const;

  const_iterator begin() const;
  const_iterator end()   const;

  // Doesn't return a null-terminated pointer, MUST pass in a pointer for len
  const char *get_by_idx(int idx, int *fid, int *val_len) const;
  RT_FidField get_by_idx(int idx)                         const;

  int  count()           const;
  bool exists(int fid)   const;
  bool empty()           const;
  bool is_stale()        const;
  bool is_permissioned() const;

  // Doesn't return a null-terminated pointer, MUST pass in a pointer for len
  const char *get_permissions(int *perms_len) const;
  RT_Field    get_permissions()               const;

  // Doesn't return a null-terminated pointer, MUST pass in a pointer for len
  const char *get_stale_reason(int *rsn_len)  const;
  RT_Field    get_stale_reason()              const;

  //
  // The rest is private (Clients should not use)
  //
  private:
    // Note: apply will only check if the keys match in debug mode
    rt_errno apply(const class RT_Message &msg, bool expand_record = true);
    void set_stale(bool is_stale, const char *stale_reason, int reason_sz);

    friend class  RT_Consumer;
    friend struct RT_ConsumerImpl;
    friend class ::RouterClient;
    friend class ::RouterServer;
    
    class  RT_FidCacheImpl   *cache;
    struct RT_RecordInternal *rec;
    class  RT_Message    *msg;

  public:
    RT_Record(class RT_FidCacheImpl *cache, struct RT_RecordInternal *rec) :
       cache(cache), rec(rec), msg(NULL) {}
    RT_Record() :
       cache(NULL), rec(NULL), msg(NULL) {}
    void clear();
    // returns -1 if the field does not exist (and can't be created)
    // returns  0 if the field was present (or created) and the new value was set
    // returns  1 if the new value is the same as that of the cached field
    int check_and_set(int fid, const char *value, int value_size);
    int check_and_set(int fid, const RT_Field &fld);
    // causes the record to point to a new record specified by key
    bool reset(const char *key, int key_sz);
};

//
// Rest is implementation
//
inline rt_errno RT_Record::set(int fid, const RT_Field &fld)
{ return (fld.ptr) ? set(fid, fld.ptr, fld.size) : RT_NO_ERROR; }

inline rt_errno RT_Record::set(const RT_FidField &f)
{ return (f.data.ptr) ? set(f.fid, f.data.ptr, f.data.size) : RT_NO_ERROR; }

inline int RT_Record::check_and_set(int fid, const RT_Field &fld)
{ return (fld.ptr) ? check_and_set(fid, fld.ptr, fld.size) : 
                     (exists(fid) ? 1 : -1);
}

inline bool RT_Record::empty() const
{ return (!rec && !msg) || count() == 0; }

inline RT_Field RT_Record::get_key() const
{ RT_Field f; f.ptr = get_key(&f.size); return f; }

inline RT_FidField RT_Record::get_by_idx(int idx) const
{ RT_FidField f; f.data.ptr = get_by_idx(idx, &f.fid, &f.data.size); return f; }

inline RT_Field RT_Record::get_permissions() const
{ RT_Field f; f.ptr = get_permissions(&f.size); return f; }

inline RT_Field RT_Record::get_stale_reason() const
{ RT_Field f; f.ptr = get_stale_reason(&f.size); return f; }

inline RT_Field RT_Record::get(int fid) const
{  RT_Field f; f.ptr = get(fid, &f.size); return f; }

class RT_Record::const_iterator
{
  public:
  typedef RT_FidField value_type;

  const_iterator() : rec(NULL), idx(255) {}
  const_iterator(const RT_Record *r, int i) : rec(r), idx(i) { fetch(); }
  bool operator!=(const const_iterator &rhs) const { return rhs.idx != idx; }
  bool operator==(const const_iterator &rhs) const { return rhs.idx == idx; }

  const_iterator operator++()
  { ++idx; fetch(); return *this; }

  const_iterator operator++(int)
  { const_iterator tmp(*this); ++idx; fetch(); return tmp; }

  const value_type &operator*()  const { return field; }
  const value_type *operator->() const { return &field; }

  private:
  const RT_Record *rec;
  value_type  field;
  int         idx;
  void fetch()
  { field.data.ptr = rec->get_by_idx(idx, &field.fid, &field.data.size); }
};

inline RT_Record::const_iterator RT_Record::begin() const
{ return const_iterator(this, 0);       }
inline RT_Record::const_iterator RT_Record::end()   const
{ return const_iterator(this, count()); }

static inline std::ostream& operator<<(std::ostream &os, const RT_Record &rec)
{
  int ky_sz;
  const char *ky = rec.get_key(&ky_sz);
  os << "Key: "; os.write(ky, ky_sz);
  os << " (NumFids = " << rec.count() << ")\n";
  for (RT_Record::const_iterator it=rec.begin(); it != rec.end(); ++it)
     os << "     " << *it << '\n';
  return os;
}

} // namespace

#endif
