#ifndef __RTLIB_API_H_20051201__
#define __RTLIB_API_H_20051201__

#include <cstddef>
#include <string>
#include <sstream>
#include <fire/private/prefix.h>
#include <fire/detail/RT_Request.h>
#include <fire/detail/RT_Message.h>
#include <fire/detail/RT_Record.h>
#include <fire/detail/RT_FieldMap.h>
#include <fire/detail/RT_Consumer.h>

namespace FIRE_NAMESPACE {

enum RT_LogLevel {
  RT_LOG_ALL   =  0,
  RT_LOG_DEBUG =  1,
  RT_LOG_INFO  =  2,
  RT_LOG_WARN  =  3,
  RT_LOG_ERROR =  4,
  RT_LOG_PANIC =  5,
  RT_LOG_NONE  =  6
};

class RT_API FDF
{
public:
  //
  // Copied over from RT_Consumer for reference...
  //
  // typedef void (*CtrlCB)(bool connected, const RT_Message *msg, void *closure);
  // typedef void (*MesgCB)(int tag, const class RT_Message  *msg,
  //                                 const class RT_Record   *rec, void *closure);
  // typedef void (*NtfyFxn)(void *closure);

  //
  // Logging functions
  //
  static rt_errno log_open(const char *file, bool append = false);
  static void     log_close();
  static void     log(RT_LogLevel severity, const char *msg, size_t size);

  // default level is RT_LOG_ALL meaning everything is logged.  All levels
  // greater than or equal to the current log level will get logged.  Use
  // RT_LOG_NONE to stop all logging.   The function returns the previous level.
  static RT_LogLevel log_level(RT_LogLevel new_level);
  static size_t      log_set_max(size_t new_max_limit);

  //
  // Property Management (for Configuration).
  //
  static const size_t MAX_PROPERTY_SIZE = 256;  // including NULL termination

  static size_t   get_property(const char *name, char *dest, 
                               const char *def_val = NULL);
  static void     set_property(const char *name, const char *value);
  static rt_errno load_properties(const char *uri,  bool append = false);

  //
  // Methods that call the default RT_Consumer
  //
  static void        set_connection_ctrl_cb(RT_Consumer::CtrlCB cb, void *closure,
                                            void **old_closure = NULL);
  static const char *get_connection_info(size_t *dest_sz = NULL);
  static rt_errno    set_connection_info(const char *connection_string);
  static rt_errno    set_connection_info(const char *host_port,
                                         const char *user,
                                         const char *passwd);

  static rt_errno set_connection_info(const char* host_port,
                                      const char* user,
                                      const char* device_id,
                                      const char* key,
                                      const char* counter,
                                      const char* path            = NULL,
                                      bool        force_input     = false,
                                      bool        machine_account = true);

  static rt_errno set_otp_app_id(const char* app_id);

  static rt_errno request_passcode(const char* user, const char* serial);

  static rt_errno get_key_and_counter(const char* user,
                                      const char* serial,
                                      const char* passcode,
                                      KeyCounter& key_counter_out);

  static rt_errno connect(bool async_connect = false);
  static rt_errno disconnect(bool keep_registrations = false);
  static bool     is_connected();
  static const char *get_connected_host();

  static rt_errno workstation_connect(const char *user, const char *passwd = "",
                                      bool async_connect = false);

  static rt_errno request(const class RT_Request &req, RT_Consumer::MesgCB cb, 
                          void *closure, int *tag);
  static rt_errno request(const class RT_Request &req, RT_Consumer::MesgCB cb, 
                          void *closure, std::vector<int> &tags);
  static rt_errno cancel(int tag, void **closure = NULL);
  static rt_errno dispatch(int max_wait_time_in_ms = 0);

  static rt_errno lock();
  static rt_errno unlock();

  static rt_errno get_record(int tag, const RT_Record **r, void **closure = NULL);
  static size_t   get_services(RT_Message *msg);
  static rt_errno update_cb(int tag, RT_Consumer::MesgCB cb, void *closure,
                                                    void **old_closure = NULL);
  static unsigned int num_tags(unsigned int *srv_tags);
  static int          get_next_tag(int tag_after, unsigned int *srv_tag = NULL);

 
  // Functions to help applications manage the event loop
  static RT_Consumer::NtfyFxn set_notify_function(RT_Consumer::NtfyFxn notify_fxn,
                               void *closure,
                               RT_Consumer::NtfyFxn thread_create_fxn = NULL,
                               RT_Consumer::NtfyFxn thread_destroy_fxn = NULL,
                               void **old_closure = NULL);
  static rt_errno get_notify_socket(void *socket);
  static rt_errno create_mswin_dispatch_window();
  static rt_errno destroy_mswin_dispatch_window();

  //  plug-ins
  static bool enable(RT_Feature &ftr);
  static bool disable(RT_Feature &ftr);
  static bool is_active(RT_Feature &ftr);

  // unchanged fields
  static void set_send_unchanged_fields(bool enable);

  // http proxy
  static bool set_http_proxy(const char *proxy_host);
  static bool set_http_proxy(const char *proxy_host,
                             const char *proxy_user,
                             const char *proxy_password);
};

#include <fire/private/rt_api.hpp>

} // namespace

//
// LOG Macros that applications may use
//

#define RT_LOG_RAW(msg)   do { std::ostringstream l_ostr19; l_ostr19 << msg; FIRE_NAMESPACE::rt_log_p(FIRE_NAMESPACE::RT_LOG_ALL, l_ostr19.str().data(), l_ostr19.str().size(), __FILE__, __LINE__); } while(0)

#define RT_LOG_INFO(msg)  do { std::ostringstream l_ostr19; l_ostr19 << msg << '\n';  FIRE_NAMESPACE::rt_log_p(FIRE_NAMESPACE::RT_LOG_INFO, l_ostr19.str().data(), l_ostr19.str().size(), __FILE__, __LINE__); } while(0)

#define RT_LOG_WARN(msg)  do { std::ostringstream l_ostr19; l_ostr19 << msg << '\n';  FIRE_NAMESPACE::rt_log_p(FIRE_NAMESPACE::RT_LOG_WARN, l_ostr19.str().data(), l_ostr19.str().size(), __FILE__, __LINE__); } while(0)

#define RT_LOG_ERROR(msg) do { std::ostringstream l_ostr19; l_ostr19 << msg << '\n';  FIRE_NAMESPACE::rt_log_p(FIRE_NAMESPACE::RT_LOG_ERROR, l_ostr19.str().data(), l_ostr19.str().size(), __FILE__, __LINE__); } while(0)

#define RT_LOG_PANIC(msg) do { std::ostringstream l_ostr19; l_ostr19 << msg << '\n';  FIRE_NAMESPACE::rt_log_p(FIRE_NAMESPACE::RT_LOG_PANIC, l_ostr19.str().data(), l_ostr19.str().size(), __FILE__, __LINE__); } while(0)
#endif
