#include <iostream>
#include <string>
#include <cstring>
#include <cstdlib>
#include <fstream>
#include <errno.h>
#include <vector>
#include <assert.h>
#include <time.h>
#include "FDS/rt_api.h"
#include "FDS/rt_fields.h"

#ifdef WIN32
// Windows doesn't like (void *) casts :<
#pragma warning(disable:4311)
#pragma warning(disable:4312)
#endif

using namespace std;
using namespace FDS;

// Uppercase is always tougher to type :>
#define log_info  RT_LOG_INFO
#define log_error RT_LOG_ERROR
#define log_warn  RT_LOG_WARN

// Configuration filename.  Can be changed via the first argument to main()
#define CONFIGURATION_FILE "mflow.cfg"

// 
// Internal represenation of an Equity Record
//
struct StockRecord
{
  StockRecord() : tag(-1), stale(false), ntrades(0), nquotes(0), nupdates(0),
                  bid(0.0), ask(0.0), last(0.0), mid(0.0),
                  trade_buy_flow(0.0), trade_sell_flow(0.0),
                  quote_buy_flow(0.0), quote_sell_flow(0.0), latest_turnover(0.0),
                  upticks(0), dnticks(0), eqticks(0) {}

  int          tag;       // The API subscription tag
  bool         stale;     // subscription stale flag.
  string       symbol;    // Symbol in the the file.  It can have a service prefix
  string       req_key;   // Requested key (e.g. can be a CUSIP, ISIN, or SEDOL)
  string       act_key;   // Actual key (i.e. key returned from 1st message)

  unsigned int ntrades;   // number of trades that were received
  unsigned int nquotes;   // number of quotes that were received
  unsigned int nupdates;  // number of other updates (not trades or quotes)

  double bid;             // Current bid
  double ask;             // Current ask
  double last;            // Last traded price
  double mid;             // Mean of the bid and ask (a.k.a midprice)
  double prev_close;      // Yesterday's closing price

  double trade_buy_flow;  // Positive money-flow (using last tick direction)
  double trade_sell_flow; // Negative money-flow (using last tick direction)
  double quote_buy_flow;  // Positive money-flow (using mid price tick direction)
  double quote_sell_flow; // Negative money-flow (using mid price tick direction)
  double latest_turnover; // The turnover between two mid price changes

  unsigned int upticks;   // Number of true upticks    (last > prev_last_price)
  unsigned int dnticks;   // Number of true downticks  (last < prev_last_price)
  unsigned int eqticks;   // Number of unchanged ticks (last == prev_last_price)
};

// Prints a StockRecord object
ostream &operator<<(ostream &os, const StockRecord &rec)
{
  os << rec.tag << ": " << rec.symbol << " last: " << rec.last << "bid: "
     << rec.bid << " ask: " << rec.ask << " mid: " << rec.mid  << " chg: "
     << rec.last - rec.prev_close << '\n'
     << "   TradeFlow = " << rec.trade_buy_flow - rec.trade_sell_flow
     << " QuoteFlow = " << rec.quote_buy_flow - rec.quote_sell_flow
     << " Tks: " << rec.upticks << "/" << rec.dnticks << "/" << rec.eqticks
     << " Net: " << (int)(rec.upticks - rec.dnticks);
  return os;
}

// Just use a simple vector of records to store the pricing data.  (Reason: The
// program just goes through the records and dumps them to a CSV file).
// If lookups by symbol are needed, the vector should either be sorted by ticker
// (so a binary search could be employed). However, a better solution would be to
// use a HashMap.  Note: We technically do a lookup by symbol when we get a
// update from the FactSet Data Server.  However, we can use our closure argument
// to locate the correct record quickly.
typedef   vector<StockRecord> PriceFile;
typedef   PriceFile::iterator PFIter;

// For now just declare a global object to store the prices
PriceFile price_file;

//
// Function to read in the symbol file
//
int read_symbol_file(const char *file)
{
  ifstream infile(file);
  if (!infile) {
    log_error("Unable to open symbol file: " << file << " err:" << errno);
    return (errno) ? errno : -1;
  }

  char buff[513];
  buff[512] = 0;

  while (infile.getline(buff, sizeof(buff)-1)) {
    char *pp = buff;

    // Get rid of C++ style comments.  Note: If we do this we may have to convert
    char *comment = strstr(pp, "//");
    if (comment) *comment = '\0';

    // Also skip shell style comments (first char = '#') and blank lines
    if (*pp == 0 || *pp == '#')
      continue;

    string symbol = buff;
    price_file.push_back(StockRecord());
    price_file.back().symbol = symbol;
    //log_info("got input symbol = " << symbol);
  }

  return 0;
}

//
// The callback will receive API messages via this function.
//
void on_message(int tag, const RT_Message *msg, const RT_Record *rec,
                 void *closure)
{
  unsigned long index = (unsigned long)closure;
  assert(index < price_file.size() && price_file[index].tag == tag);

  StockRecord *stock = &price_file[index];

  // first handle any errors, or closing of the update stream
  if (msg->is_closed()) {
    log_error("Error stream closed on sym: " << stock->symbol << " err:"
              << msg->get_error() << " desc: " << msg->get_error_description());
    FDF::cancel(tag);
    stock->tag   = -1;
    stock->stale = true;
    return;
  }

  // See if the stream went stale.  This case will recover on its own, so it is
  // not neccessary to cancel the stream
  if (msg->is_stale()) {
    log_warn("sym: " << stock->symbol << " stale: " << msg->get_stale_reason());
    stock->stale = true;
  }
  else {
    stock->stale = false;
  }

  if (msg->is_response()) {
    // This is the first message, set our state and just return
    stock->act_key = msg->get_key();
    stock->bid  = rec->get(FIDS::BID_1).to_double();
    stock->ask  = rec->get(FIDS::ASK_1).to_double();
    stock->mid  = (stock->bid + stock->ask) / 2;
    stock->last = rec->get(FIDS::LAST_1).to_double();
    stock->prev_close = rec->get(FIDS::PREV_CLOSE).to_double();
    return;
  }

  // It is an update.  Get our message type
  RT_Field msg_type = msg->get(FIDS::MSG_TYPE);
  if (msg_type.empty()) {
    log_error("Message missing MSG_TYPE msg:" << *msg);
    return;
  }

  if (*msg_type.ptr == 'Q') {
    // Handle quote messages
    ++stock->nquotes;
    stock->bid = rec->get(FIDS::BID_1).to_double();
    stock->ask = rec->get(FIDS::ASK_1).to_double();
    double new_mid = (stock->bid + stock->ask) / 2;
    if (new_mid > stock->mid) {
      stock->quote_buy_flow  += stock->latest_turnover;
      stock->latest_turnover  = 0.00;
    }
    else if (new_mid < stock->mid) {
      stock->quote_sell_flow += stock->latest_turnover;
      stock->latest_turnover  = 0.00;
    }
    stock->mid = new_mid;
  }

  else if (*msg_type.ptr == 'T') {
    // Handle trade messages
    ++stock->ntrades;
    unsigned int new_last_vol = rec->get(FIDS::LAST_VOL_1).to_uint();
    double       new_last     = rec->get(FIDS::LAST_1).to_double();
    if (new_last > stock->last) {
      ++stock->upticks;
      stock->trade_buy_flow  += (new_last * new_last_vol);
    }
    else if (new_last < stock->last) {
      ++stock->dnticks;
      stock->trade_sell_flow += (new_last * new_last_vol);
    }
    else {
      ++stock->eqticks;
    }
    stock->latest_turnover += (new_last * new_last_vol);
    stock->last             = new_last;
  }

  else {
    // Not a quote or a trade (i.e. summary, verify, cancel, correction, etc.)
    ++stock->nupdates;
  }
}

//
// Function to print the output file
//
void print_output_file(ostream &ofile)
{
  size_t num_syms = price_file.size();
  log_info("Writing output file for " << (unsigned int)num_syms << " symbols...");

  ofile.seekp(0);
  for (unsigned int i=0; i < num_syms; ++i)
    ofile << price_file[i] << '\n';
  ofile << "\n----------- END ---------------- " << endl;

  log_info("done");
}

int main(int argc, char* argv[])
{
  string config_file = (argv[1]) ? (argv[1]) : CONFIGURATION_FILE;

  // open our log file
  // FDF::log_open("mflow.log");

  log_info("mflow started.  Opening config file " << config_file);

  rt_errno rterr = FDF::load_properties(config_file.c_str());
  if (rterr) {
    log_error("Couldn't load properties: " << config_file
              << " e:" << rterr);
    return rterr;
  }

  // Generic buffer to get our configuration values
  char prop_val[FDF::MAX_PROPERTY_SIZE];

  // Get the field map file
  RT_FieldMap *map;
  FDF::get_property("FIELD_MAP_FILE", prop_val);
  map = RT_FieldMap::create(prop_val);
  if (!map)
    log_warn("Unable to load field map file: " << prop_val);

  // Get our symbol file (defaults to mflow.symbols)
  // Read in the symbols.  Exit program if we can't read in our tickers.
  FDF::get_property("SYMBOL_FILE", prop_val, "mflow.symbols");
  if (read_symbol_file(prop_val)) return -1;

  // Get our output timer (defaults to 5 minutes) and file
  FDF::get_property("OUTPUT_TIMER_SEC", prop_val, "300");
  unsigned int output_timer_secs = (unsigned int)atoi(prop_val);
  if (!output_timer_secs) {
    log_warn("Invalid output timer, must be non-zero val= " << output_timer_secs);
    output_timer_secs = 300;
  }

  // Output file defaults to "mflow.txt"
  ofstream out_file;
  FDF::get_property("OUTPUT_FILE", prop_val, "mflow.txt");
  out_file.open(prop_val);
  if (!out_file) {
    log_error("Unable to open outfile: " << prop_val << " e:" << errno);
    return errno;
  }
 
  // Get our data service (defaults to "FDS1")
  FDF::get_property("DATA_SERVICE", prop_val, "FDS1");
  string data_service(prop_val);

  // Get OTP details
  FDF::get_property("OTP_FULL_PATH", prop_val, "");
  string otp_full_path(prop_val);

  if(otp_full_path.empty()) {
    // Passing NULL into set_connection_info will cause it to look for the
    // RT_CONNECTION property using FDF::get_property() 
    FDF::set_connection_info(NULL);
  } else {
    // Get host details
    FDF::get_property("RT_HOST", prop_val, "");
    string host(prop_val);

    // Get user details
    FDF::get_property("RT_USER", prop_val, "");
    string user_serial(prop_val);

    string path = "";
    string device_id = "";
    size_t sep_loc = otp_full_path.find_last_of("/\\");
    if(sep_loc!=string::npos) {
      path = otp_full_path.substr(0, sep_loc);
      device_id = otp_full_path.substr(sep_loc+1);
    } else {
      device_id = otp_full_path;
    }
    size_t dot_loc = device_id.find(".");
    if(dot_loc!=string::npos) {
      device_id = device_id.substr(0, dot_loc);
    }
    log_info("OTP connection details host:" << host << " user:" << user_serial << " device_id:" << device_id << " in path:" << path);
    rterr = FDF::set_connection_info(host.c_str(), user_serial.c_str(), device_id.c_str(), NULL, NULL, path.c_str(), false);
    if(rterr) {
      log_error("error in setting otp connection info e:" << rterr);
      return rterr;
    }
  }

  // Connect to the server
  rterr = FDF::connect();
  if (rterr) {
    log_error("Unable to connect " << rterr);
    return -1;
  }

  // Make real-time requests for all symbols in our price_file
  for (unsigned long i=0; i < price_file.size(); ++i) {
    int tag;

    // Make a request.
    // Note: we are passing the array index as the closure argument
    RT_Request req(data_service.c_str(), price_file[i].symbol.c_str());
    rterr = FDF::request(req, &on_message, (void *)i, &tag);

    log_info("requested " << price_file[i].symbol << " tag=" << tag);
    if (rterr) log_warn("Request is queued e: " << rterr);

    // Set the state that we know of (like tag, and requested key)
    price_file[i].tag     = tag;
    price_file[i].req_key = req.get_key();

    // When requesting many tickers (like > 5000), it is better to dispatch
    // messages while requesting.  This way the system is not queuing up
    // messages internally (since there is an API thread reading the TCP
    // socket).  Note:  This is optional, and can be commented out if desired.
    rterr = FDF::dispatch(0);
  }

  // Initialize our timer variables
  time_t curr_time, prev_time;
  curr_time = prev_time = time(0);
  unsigned int elapsed_time = 0; 

  // Our main loop
  while (1) {
    rterr = FDF::dispatch((output_timer_secs - elapsed_time) * 1000);
    if (rterr) {
      log_warn("dispatch err: " << rterr);
    }
    curr_time = time(0);
    elapsed_time = elapsed_time + (unsigned int)(curr_time - prev_time);

    if (elapsed_time >= output_timer_secs) {
      // time to write our output file, write it and reset elapsed_time
      print_output_file(out_file);
      elapsed_time = 0;
    }

    prev_time = curr_time;
  }
}


